/**************************************************************************


           Copyright 1986 University of Massachusetts.
                         All rights reserved.

     Permission to copy and modify this software and its documen-
     tation only for internal use in your organization is hereby
     granted, provided that this notice is retained thereon and
     on all copies.  UMASS makes no representations as to the sui-
     tability and operability of this software for any purpose.
     It is provided "as is" without express or implied warranty.

     UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
     NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
     INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
     SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
     IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
     ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
     FORMANCE OF THIS SOFTWARE.

     No other rights, including, for example, the right to redis-
     tribute this software and its documentation or the right to
     prepare derivative works, are granted unless specifically
     provided in a separate license agreement.

     Copyright 1986, University of Massachusetts. All rights
     reserved.




DATE:      (development)
PROGRAM:     Fast Region-Based Line Finder (based upon Burns' line finder)
DESIGNERS:   Philip Kahn and Leslie Kitchen
PROGRAMMER:  Philip Kahn
DESCRIPTION:

Modified to use uncertainty window around expected line by James H. Burrill.
/***************************************************************************/

#include <stdio.h>
#include <math.h>
#include "sequent.h"
#include  "tableau.h"

extern int my_id;
extern MODEL_INFO *model_info;
extern int number_models;
extern int row_size;
extern int col_size;
extern int current_plane;
extern IMAGE *image;
extern slock_t model_lock;
extern int report;

static shared float *thetas;
static shared int quadrant_size;

/* line component structures */

typedef struct { /* line characteristics definition */
  /******* fitted line descriptions *******/
  /*
    the endpoints of the lines where the order of (x1,y1) and (x2,y2)
    encodes the sense of the line (right is bright)
  */
  DL endpoints;

  /* The orientation from -pi to pi which encodes the sense of the line */
  float theta;
  float length;   /* estimate of the length of the fitted line */
  float grad_mag; /* the gradient magnitude of the line (roughly, "contrast") */

  /*
    the extent to which the line fits the region; that is, the
    amount by which the line fits the region.  Calculated
    as 0 <= min(Eigenvalue)/max(Eigenvalue) <= 1.
  */
  float dispersion;
  float region;      /* the region number to which the line was fitted */
  float num_pixels_in_region; /* number of pixels in the line support region */
  float sum_delta_x; /* sum of the gradient vectors for all pixels in region */
  float sum_delta_y;
  float sum_grad_mag;
  } LINE;

/*** bucket range list data structures ***/

/*
  The orientation bounds (in radians) where the low bound
  (inclusive) is followed by the high bound (exclusive)
  which defines the range for a bucket; it is assumed that
  min_theta <= max_theta 
*/

typedef struct { /* Bucket Range (BR) definition */
  float min_theta;
  float max_theta;
  } BUCKET_RANGE;

/*** Bucket Table (bt) type declarations ***/

#define BUCKET unsigned char

/*
   The center of the bucket table is considered (0,0);
   thus, it is split into four quandrants.  It is assumed
   that the side of the table is a power of two, thus, the
   following values specify the maximum absolute value that
   can occur in any of the quadrants as well as the size of
   each side of the 2d Bucket Table array.  These values are
   initialized when the data structure is created.
*/
typedef struct {
  int quadrant_x;
  int quadrant_y;
  int size_x;
  int lookup_size;
  int byte_size;
  BUCKET lookup[1]; /* 1 .. n entries */
  } BUCKET_TABLE;

/*** Line Filter Parameters (LFP) data structure ***/

typedef struct {
  struct {
    BOOLEAN active; /* TRUE if filter is active. */
    float min_theta; /* range of acceptable line directions (in radians) */
    float max_theta;
    } direction;
  struct { /*(tentative)*/
    BOOLEAN active; /* TRUE if filter is active. */
    float min_theta; /* range of acceptable line orientations (in radians) */
    float max_theta;
    } orientation;
  struct {
    BOOLEAN min_active; /* TRUE if low-bound filter is active. */
    float min_grad_mag; /* min allowable gradient magnitude (inclusive) */
    BOOLEAN max_active; /* TRUE if high-bound filter is active. */
    float max_grad_mag; /* max allowable gradient magnitude (exclusive) */
    } gradient_magnitude;
  struct {
    BOOLEAN min_active; /* TRUE if low-bound filter is active. */
    float min_length;   /* minimum allowable line length (inclusive) */
    BOOLEAN max_active; /* TRUE if high-bound filter is active. */
    float max_length;   /* maximum allowable line length (exclusive) */
    } length;
  struct {
    BOOLEAN min_active;   /* TRUE if low-bound filter is active. */
    float min_dispersion; /* min allowable dispersion (inclusive) */
    BOOLEAN max_active;   /* TRUE if high-bound filter is active. */
    float max_dispersion; /* max allowable dispersion (exclusive) */
    } dispersion;
  } LINE_FILTER;

#define REGION_EXTANT_EXPANSION    0.5
#define GRADIENT_SCALE             3.0

#define LINK short
/*
 The following values, including the value -1, must be representable by a LINK.
*/

#define MAX_NUMBER_OF_FRAGMENTS    0x7fff
#define	MAX_NUMBER_OF_REGIONS      0x7fff
#define MAX_NUMBER_OF_EQUIVALENCES 0x7fff


/* Region Statistics Table (RST):
   These are the partial sums maintained for each region
   that are needed to fit a line and determine its parameters. 
*/
typedef struct { 
  long wsum_x2;     /* stats needed to form the Eigen scatter matrix */
  long wsum_y2;
  long wsum_xy;
  long wsum_x;
  long wsum_y;
  long sum_x;       /* stats needed to compute the centroid */
  long sum_y;
  int pixel_count;  /* number of pixels in the region */
  long sum_delta_x; /* Stats needed to compute line parameters  */
  long sum_delta_y;
  long sum_grad_mag;
  short min_x;      /* Stats needed to determine endpoints of line */
  short max_x;
  short min_y;
  short max_y;
  } RST_ENTRY;

typedef struct {
  int pixel_count; /* fragment pixel count */
  LINK fet_link;  /* index into the Fragment Equivalence Table */
  LINK region;    /* region number - NIL, UNDEFINED, number*/
  } RFT_ENTRY;

typedef struct {
  LINK frag_num; /* fragment # which is an index into the RFT */
  LINK link;     /* an FET index for the next equivalent fragment */
  } FET_ENTRY;

#define MAX_FET_FREE_PTR_VALUE MAX_NUMBER_OF_EQUIVALENCES-1

/* calculate the partial sums, assemble the Eigen scatter matrix,
   and calculate the Eigenvectors to fit a line to a region.
  
   Briefly, partial sums in the form of an Eigen scatter matrix are used
   to fit a line to a cluster of points within a region, where the scatter
   matrix is of the form [a b].  More formally, if E stands for sigma, the
     [b c]
   the traditional Eigen technique scatter matrix is 
     [  E(X^2*w) E(X*Y*w) ]
     [  E(X*Y*w) E(Y^2*w) ], where w is a weighting factor (e.g., 
   gradient magnitude); yet, this assumes that all (X,Y) pairs have their 
   origins relocated to the centroid of the original scatter points; this 
   leads to the following reformulation which allows all sums to be 
   accumulated in a single pass (where (X,Y) is a point whose origin is 
   the centroid; (x,y) is a point retaining its original origin; and n is 
   the number of (x,y) pairs in the region:
    E(X^2*w) = E(x^2*w) - [E(x*w)^2]      / E(w)
    E(X*Y*w) = E(x*y*w) - [E(x*w)*E(y*w)] / E(w)
    E(Y^2*w) = E(y^2*w) - [E(y*w)^2]      / E(w)
   Note that x and y are weighted by their associated gradient magnitude; 
   this incorporates the assumption that high gradient magnitudes are
   better indicators of an underlying line component.
*/

#define ODD(value) ((value) & 1)

/* This macro takes a floating radian angle and brings it into the range
/* of -PI to PI. */
#define MAKE_THETA_IN_RANGE(theta) \
{\
REAL tmp_float;\
int num_PI;\
/* bring theta into range */\
if (theta < 0.0) {\
 tmp_float = (-theta) / PI;\
 num_PI = (int) tmp_float;\
 theta = theta + (ODD(num_PI) ? (ceil(tmp_float) * PI) : (num_PI * PI));\
 }\
else { /* theta >= 0.0 */\
 tmp_float = theta / PI;\
 num_PI = (int) tmp_float;\
 theta = theta - (ODD(num_PI) ? (ceil(tmp_float) * PI) : (num_PI * PI));\
 }\
}

BUCKET_TABLE *alloc_bt()
{
int x_size, lookup_size, byte_size;
BUCKET_TABLE *new_bt;

  /* verify that the quadrant size is a power of 2 */

  lookup_size = quadrant_size;
  while (lookup_size > 1) {
    if (lookup_size & 1) return NULL;
    lookup_size >>= 1;
    }

  /* compute the bucket table size in bytes */

  x_size = (quadrant_size * 2) + 1; /* 1 is for the origin */
  lookup_size = x_size * x_size; /* it's a 2d table */

  /*** create the table ***/

  byte_size = sizeof(BUCKET_TABLE) + /* table info */
                (sizeof(BUCKET) * lookup_size); /* table data */
  new_bt = (BUCKET_TABLE *) malloc(byte_size);
  if (new_bt != NULL) { /* initialize the structure */
    new_bt->size_x = x_size;
    new_bt->byte_size = byte_size;
    new_bt->lookup_size = lookup_size;
    new_bt->quadrant_x = quadrant_size;
    new_bt->quadrant_y = quadrant_size;
    }
  return new_bt;
}

void setup_for_buckets(quad_size)
int quad_size;
{
int x_size, x, y;
float *thetap;

  quadrant_size = quad_size;
  x_size = (quadrant_size * 2) + 1; /* 1 is for the origin */
  thetas = (float *) malloc(x_size * x_size * sizeof(float));
  if (thetas == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate thetas!\n");
    exit(1);
    }

  /* compute the radian angle for the lookup table position */

  thetap = thetas;
  for (x = -quadrant_size; x <= quadrant_size; x++)
    for (y = -quadrant_size; y <= quadrant_size; y++)
      *thetap++ = ((x == 0) && (y == 0)) ?
                        0.0 : atan2((double) y, (double) x);
}

/*
  This function takes a Bucket Range List (brl and translates
  it into a two dimensional lookup table where each axis is the
  magnitude at a pixel in the x and y directions. This is roughly
  equivalent to computing arctan[mag_y/mag_x] within some range 
  and then classifying the "bucket".
   The building of this table avoids the computation of the
  arctan, a division, and the control overhead needed to classify
  the result into a sector "bucket." Note that it causes
  nonlinearities due to using a lookup table for the arctan 
  (i.e., all positions in the table do not possess the same 
  range of arctan coarseness).
  ALGORITHM:
  This is the dumb algorithm which computes atan(delta_y/delta_x)
  and then coarse codes it into the bucket table.

  bucket_range_list IN: a list of line orientation ranges (in radians)
  bucket_table INOUT: region_range_list encoded into this lookup table
*/

int build_bt(bucket_range_list,number_ranges,bucket_table)
BUCKET_RANGE *bucket_range_list;
int number_ranges;
BUCKET_TABLE *bucket_table;
{
BUCKET bucket_num;
int x, y, size_x, size_y;
int brl_ptr;
REAL bucket_theta;
int index;

  /*** use the Bucket Range List to fill the Bucket Table ***/

  size_x = bucket_table->quadrant_x;
  size_y = bucket_table->quadrant_y;

  for (x = -size_x; x <= size_x; x++) {
    index = size_y + ((x + size_x) * bucket_table->size_x);
    for (y = -size_y; y <= size_y; y++) {

      /* Get the radian angle for the lookup table position */

      bucket_theta = *(thetas + y + index);

      /* see if a bucket has been defined for the computed theta */

      brl_ptr = 0;
      bucket_num = UNDEFINED;

      do { /*** see if the lookup theta is in a bucket ***/
        if (bucket_range_list[brl_ptr].min_theta >
            bucket_range_list[brl_ptr].max_theta) {
          /* then the bucket crosses the -PI/PI boundary */
          if ((bucket_theta >= bucket_range_list[brl_ptr].min_theta) ||
              (bucket_theta <  bucket_range_list[brl_ptr].max_theta))
            bucket_num = brl_ptr + 1;
          }
        else { /* the bucket doesn't cross the -PI/PI boundary */
          if ((bucket_theta >= bucket_range_list[brl_ptr].min_theta) &&
              (bucket_theta <  bucket_range_list[brl_ptr].max_theta))
            bucket_num = brl_ptr + 1;
          }
        } while ((++brl_ptr < number_ranges) && (bucket_num == UNDEFINED));

      /* store into the bucket lookup table */
      /* set the bucket lookup table location */

      bucket_table->lookup[index + y] = bucket_num;
      }
    }

  return 0;
}

/**************************************************************************/

/*
  This code takes the signed theta and centroid for a line and the extant 
  of the region to then compute the endpoints of the line; that is, it 
  computes the two points where the line, denoted by theta and the centroid, 
  intersects the bounding extant box.
*/

int compute_endpoints(x_min, y_min, x_max, y_max,
                       theta, centroid_x, centroid_y,
                       nx1, ny1, nx2, ny2)
REAL x_min, x_max, y_min, y_max;
REAL theta;
REAL centroid_x, centroid_y;
REAL *nx1, *ny1, *nx2, *ny2;
{
REAL sigma, tan_sigma;
REAL adjacent, opposite;
REAL centroid_to_y;
REAL x1, x2, y1, y2;

  /* determine which quadrant theta is in */
  /* find the intersections of the line and box */

  if (theta >= 0.0) {
    if (theta < PI_OVER_2) { /* Quadrant 1 */

    sigma = theta;
    tan_sigma = tan(sigma);

      /* compute (x1,y1) */

      adjacent = centroid_x - x_min;
      opposite = tan_sigma * adjacent;
      centroid_to_y = centroid_y - y_min;
      if (opposite > centroid_to_y) { /* line goes through bottom of box */
        y1 = y_min;
        x1 = centroid_x - (centroid_to_y / tan_sigma);
        }
      else { /* line goes through left side of box */
        x1 = x_min;
        y1 = centroid_y - opposite;
        }

      /* compute (x2,y2) */

      adjacent = x_max - centroid_x;
      opposite = tan_sigma * adjacent;
      centroid_to_y = y_max - centroid_y;
      if (opposite > centroid_to_y) { /* line goes through top of box */
        y2 = y_max;
        x2 = centroid_x + (centroid_to_y / tan_sigma);
        }
      else { /* line goes through right side of box */
        x2 = x_max;
        y2 = centroid_y + opposite;
        }
      }

    else if (theta > PI_OVER_2) { /* Quadrant 2 */
      sigma = PI - theta;
      tan_sigma = tan(sigma);

      /* compute (x1,y1) */

      adjacent = x_max - centroid_x;
      opposite = tan_sigma * adjacent;
      centroid_to_y = centroid_y - y_min;
      if (opposite > centroid_to_y) { /* line goes through bottom of box */
        y1 = y_min;
        x1 = centroid_x + (centroid_to_y / tan_sigma);
        }
      else { /* line goes through right side of box */
        x1 = x_max;
        y1 = centroid_y - opposite;
        }

      /* compute (x2,y2) */

      adjacent = centroid_x - x_min;
      opposite = tan_sigma * adjacent;
      centroid_to_y = y_max - centroid_y;
      if (opposite > centroid_to_y) { /* line goes through top of box */
        y2 = y_max;
        x2 = centroid_x - (centroid_to_y / tan_sigma);
        }
      else { /* line goes through left side of box */
        x2 = x_min;
        y2 = centroid_y + opposite;
        }
      }
    else { /* vertical line */
      y1 = y_min;
      x1 = centroid_x;

      /* compute (x2,y2) */

      y2 = y_max;
      x2 = centroid_x;
      }
    }

  else {
    if (theta < -PI_OVER_2) { /* Quadrant 3 */
      sigma = PI + theta;
      tan_sigma = tan(sigma);

      /* compute (x1,y1) */

      adjacent = x_max - centroid_x;
      opposite = tan_sigma * adjacent;
      centroid_to_y = y_max - centroid_y;

      if (opposite > centroid_to_y) { /* line goes through top of box */
        y1 = y_max;
        x1 = centroid_x + (centroid_to_y / tan_sigma);
        }
      else { /* line goes through right side of box */
        x1 = x_max;
        y1 = centroid_y + opposite;
        }

      /* compute (x2,y2) */

      adjacent = centroid_x - x_min;
      opposite = tan_sigma * adjacent;
      centroid_to_y = centroid_y - y_min;

      if (opposite > centroid_to_y) { /* line goes through bottom of box */
        y2 = y_min;
        x2 = centroid_x - (centroid_to_y / tan_sigma);
        }
      else { /* line goes through left side of box */
        x2 = x_min;
        y2 = centroid_y - opposite;
        }
      }
    else if (theta > -PI_OVER_2) { /* Quadrant 4 */
      sigma = (-theta);
      tan_sigma = tan(sigma);

      /* compute (x1,y1) */

      adjacent = centroid_x - x_min;
      opposite = tan_sigma * adjacent;
      centroid_to_y = y_max - centroid_y;

      if (opposite > centroid_to_y) { /* line goes through top of box */
        y1 = y_max;
        x1 = centroid_x - (centroid_to_y / tan_sigma);
        }
      else { /* line goes through left side of box */
        x1 = x_min;
        y1 = centroid_y + opposite;
        }

      /* compute (x2,y2) */

      adjacent = x_max - centroid_x;
      opposite = tan_sigma * adjacent;
      centroid_to_y = centroid_y - y_min;

      if (opposite > centroid_to_y) { /* line goes through bottom of box */
        y2 = y_min;
        x2 = centroid_x + (centroid_to_y / tan_sigma);
        }
      else { /* line goes through right side of box */
        x2 = x_max;
        y2 = centroid_y - opposite;
        }
      }
    else  { /* vertical line */
      y1 = y_max;
      x1 = centroid_x;

      /* compute (x2,y2) */

      y2 = y_min;
      x2 = centroid_x;
      }
    }
  if (y1 < y_min) y1 = y_min;
  if (y1 > y_max) y1 = y_max;
  if (y2 < y_min) y2 = y_min;
  if (y2 > y_max) y2 = y_max;

  if (x1 < x_min) x1 = x_min;
  if (x1 > x_max) x1 = x_max;
  if (x2 < x_min) x2 = x_min;
  if (x2 > x_max) x2 = x_max;

  *nx1 = x1;
  *nx2 = x2;
  *ny1 = y1;
  *ny2 = y2;
}

/*************************************************************************/

/* Connected Components Algorithm (CCA):
    Two tables are used specifically by the CCA.  The 
    Region Fragment Table (RFT) has an entry for 
    every fragment found in the image.  After PASS 1,
    it contains a fragment pixel count and a pointer into
    the Fragment Equivalence Table (FET); after PASS 2,
    the RFT contains the region number and pointer into
    the Region Statistics Table (RST).  The Fragment 
    Equivalence Table (FET) contains information needed 
    to merge fragments into regions.
*/

/*
  bucket_table  IN: table to classify pixels into orientation buckets
  filters       IN: filters for controlling the output of this routine
  out_line_list INOUT: if successfully executed, then a list of extracted
                       lines is output; otherwise undefined.
  RLI           INOUT: if RLI != NULL and successfully 
                       executed, then Region Labelled Image (RLI) is created;
                       otherwise, it is undefined.
*/

int line_finder(image, image_rows, image_cols, display,
               x1, y1, x2, y2, cx, cy,
               bucket_table, out_line_list, number_lines,
               RLI,line_filters, pixel_count_threshold,grad_mag_threshold)
IMAGE *image;
int image_rows;
int image_cols;
int display;
REAL x1, y1, x2, y2, cx, cy;
BUCKET_TABLE *bucket_table;
LINE *out_line_list;
int number_lines; /* Number of lines allowed in out_line_list */
RLI_TYPE *RLI;
LINE_FILTER *line_filters;
int pixel_count_threshold;
REAL grad_mag_threshold;
{
short *gradient_image; /* pixel gradient magnitude computed in PASS 1 */
short *delta_x_image;  /* pixel gradient, x direction, computed in PASS 1 */
short *delta_y_image;  /* pixel gradient, y direction, computed in PASS 1 */
LINK *FLI;            /* Fragment Label Image created in PASS 1 */

/*
   neighborhood bucket values needed by the CCA (where the
   neighborhood is A where C is the current pixel's bucket)
     BC
*/
BUCKET *prev_row_bucket;   /* for A */
BUCKET prev_pixel_bucket; /* for B */
BUCKET cur_bucket;        /* current pixel's bucket # */

RST_ENTRY *rst_ptr;
RFT_ENTRY *rft_ptr;
RFT_ENTRY *frag_dest;
RFT_ENTRY *frag_merger;
FET_ENTRY *fet_ptr;
FET_ENTRY *FET;
RST_ENTRY *RST;
RFT_ENTRY *RFT;

LINE new_line;

RLI_TYPE *RLI_address;
IMAGE *row1_addr, *row2_addr, *row3_addr;

int num_lines;
int crosses_discontinuity;
int num_frags;          /* most recently allocated fragment # */
int fragment;           /* current pixel's (i.e., C's) fragment # */
int num_regions;        /* number of merged fragment regions */
int current_region;     /* current pixel's region number */
int row, col;           /* current image pixel position */
int delta_x, delta_y;   /* dy/dI and dx/dI at a pixel */
int head_fet_index;     /* first node in the FET equiv. list */
int fet_free_ptr;       /* current free spot in the FET */
int rst_index;          /* current index into the RST table */
int region_pixel_count; /* # pixels in the current region */
int rft_index;
int row_reloc, col_reloc; /* offset so origin is in image center */
int sub_col, mid_col, add_col;
int loop, limit;
int frag_A, frag_B;
int row_index;            /* Into source image */
int row_offset;           /* Into temp images */
int rel_row, rel_col;
int fet_index, fet_head;
int row_count;
int col_count;
long w_x, w_y;  
REAL grad_mag;
REAL x_diff, y_diff;
REAL qe1, qe2;
REAL scat_xx, scat_xy, scat_yy; /* scatter matrix values */
REAL eivalsmall, eivecsmall, eivallarge;
REAL centroid_x, centroid_y; /* (x,y) region centroid */
REAL line_direction;
REAL x_min, y_min, x_max, y_max; 
REAL sum_x, sum_y;
REAL num_pixs;
VEC2D pt[8];
int i, j, k, flg, flp, y;
REAL minr, maxr;
REAL minc, maxc;
int minr_int, maxr_int;
int minc_int, maxc_int;
VEC2D a[4];
REAL slope[4];
REAL cc1, cc2, cc3, cc4, t;

  /* Determine oriented extents polygon */

  flp = 0;
  flg = 0;
  if      (x1 <= x2 && y1 <= y2) {
    flg = 8 + 4 + 1;
    flp = 0;
    }
  else if (x2 <= x1 && y1 <= y2) {
    flg = 8 + 4 + 2;
    flp = 0;
    }
  else if (x2 <= x1 && y2 <= y1) {
    flg = 8 +  4 + 1;
    flp = 1;
    }
  else if (x1 <= x2 && y2 <= y1) {
    flg = 4 + 2 + 8;
    flp = 1;
    }

  if (flp) {
    i = x1;
    x1 = x2;
    x2 = i;
    i = y1;
    y1 = y2;
    y2 = i;
    }

  i = 0;

  if (flg == 0) return;

  if (flg & 1) {
    pt[i].y = y1 + cy;
    pt[i++].x = x1 - cx;
    }

  if (flg & 8) {
    pt[i].y = y1 - cy;
    pt[i++].x = x1 - cx;
    }

  if (flg & 4) {
    pt[i].y = y1 - cy;
    pt[i++].x = x1 + cx;
    }

  if (flg & 2) {
    pt[i].y = y1 + cy;
    pt[i++].x = x1 + cx;
    }

  if (flg & 1) {
    pt[i].y = y2 - cy;
    pt[i++].x = x2 + cx;
    }

  if (flg & 8) {
    pt[i].y = y2 + cy;
    pt[i++].x = x2 + cx;
    }

  if (flg & 4) {
    pt[i].y = y2 + cy;
    pt[i++].x = x2 - cx;
    }

  if (flg & 2) {
    pt[i].y = y2 - cy;
    pt[i++].x = x2 - cx;
    }

  minr = pt[0].y;
  maxr = pt[0].y;
  minc = pt[0].x;
  maxc = pt[0].x;
  for (i = 1; i < 6; i++) {
    if (pt[i].y < minr) minr = pt[i].y;
    if (pt[i].y > maxr) maxr = pt[i].y;
    if (pt[i].x < minc) minc = pt[i].x;
    if (pt[i].x > maxc) maxc = pt[i].x;
    }

  minr_int = ceil(minr);
  maxr_int = floor(maxr);

  minc_int = ceil(minc);
  maxc_int = floor(maxc);

  /*  If any parameter is out of bounds, set to limit. */

  k = 0;
  for (i = 0; i < 6; i++) {
    j = i + 1;
    if (j >= 6) j = 0;
    if (display) draw_line_rgb(display - 1,pt[i].y,pt[i].x,pt[j].y,pt[j].x,BLUE);
    if (pt[i].y == pt[j].y) continue;

    slope[k] = ((double) (pt[i].x - pt[j].x)) /
                ((double) (pt[i].y - pt[j].y));
    a[k] = pt[i];
    k++;
    }

  /* Determine problem size parameters */

  if (minr_int < 1) minr_int = 1;
  if (maxr_int >= image_rows - 1) maxr_int = image_rows - 2;
  if (minc_int < 1) minc_int = 1;
  if (maxc_int >= image_cols - 1) maxc_int = image_cols - 2;

  row_count = maxr_int - minr_int + 1;
  col_count = maxc_int - minc_int + 1;

  crosses_discontinuity = (line_filters->direction.max_theta < 
                           line_filters->direction.min_theta);

  limit = row_count * col_count;

  FLI = (LINK *) malloc(limit*sizeof(LINK));
  if (FLI == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate FLI\n");
    return 0;
    }
  for (loop = 0; loop < limit; loop++) *(FLI + loop) = UNDEFINED;

  gradient_image = (short *) malloc(limit*sizeof(short));
  if (gradient_image == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate gradient_image\n");
    return 0;
    }
  delta_x_image = (short *) malloc(limit*sizeof(short));
  if (delta_x_image == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate delta_x_image\n");
    return 0;
    }
  delta_y_image = (short *) malloc(limit*sizeof(short));
  if (delta_y_image == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate delta_y_image\n");
    return 0;
    }
  prev_row_bucket = (BUCKET *) malloc(col_count * sizeof(BUCKET));
  if (prev_row_bucket == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate prev_row_bucket\n");
    return 0;
    }
  FET = (FET_ENTRY *) malloc(MAX_NUMBER_OF_EQUIVALENCES * sizeof(FET_ENTRY));
  if (FET == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate FET\n");
    return 0;
    }
  RFT = (RFT_ENTRY *) malloc(MAX_NUMBER_OF_FRAGMENTS * sizeof(RFT_ENTRY));
  if (RFT == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate RFT\n");
    return 0;
    }

  /***** verify correctness of input parameters *****/

  /***** initialize local data structures *****/

  /* offset so origin is in image center */

  row_reloc = image_rows >> 1;
  col_reloc = image_cols >> 1;

  /* incremented prior to use */

  num_frags = 0;

  /* always points to a free location */

  fet_free_ptr = 0;

 /* A in CCA is a boundary condition for the first row */

  for (col = 0; col < col_count; col++) prev_row_bucket[col] = UNDEFINED;

  /***** initialize outgoing data structures *****/

  /* incremented prior to use */

  num_lines = 0;

  /* if a Region Label Image (RLI) is to be produced, initialize
  /* its values to UNASSIGNED labels so that only assigned
  /* region labels need to be placed in the RLI  */

  if (RLI != NULL) {

    /* initialize the RLI image the fastest way possible */

    RLI_address = RLI;

    for (loop = 0; loop < limit; loop++) *RLI_address++ = UNDEFINED;
    }

  if (report > 1) fprintf(stderr,"Pass 1\n");

  /*
    PASS 1 raster scans through the intensity image.  It
     does the following:
     -- computes pixel gradient vector
     -- finds the pixel "bucket" number for that pixel
     -- eliminates pixels not of interest
     -- groups fragments using the Connected Components
        Algorithm (CCA)
     -- defines regions as fragments by building a Fragment
        Equivalence Table (FET)
  */

  if (k == 2) {
    if (a[0].x < a[1].x) {
      cc2 = a[0].x;
      cc3 = a[1].x;
      }
    else {
      cc2 = a[1].x;
      cc3 = a[0].x;
      }
    }

  for (row = 0; row < row_count; row++) {

    y = row + minr_int;

    if (k != 2) {
      cc1 = a[0].x + slope[0] * (y - a[0].y);
      cc2 = a[1].x + slope[1] * (y - a[1].y);
      cc3 = a[2].x + slope[0] * (y - a[2].y);
      cc4 = a[3].x + slope[1] * (y - a[3].y);

      /* Sort the four intersects to get the two in the middle. */

      if (cc1 > cc2) { t = cc1; cc1 = cc2; cc2 = t; }
      if (cc1 > cc3) { t = cc1; cc1 = cc3; cc3 = t; }
      if (cc1 > cc4) { t = cc1; cc1 = cc4; cc4 = t; }
      if (cc2 > cc3) { t = cc2; cc2 = cc3; cc3 = t; }
      if (cc2 > cc4) { t = cc2; cc2 = cc4; cc4 = t; }
      if (cc3 > cc4) cc3 = cc4;
      }

    row_index = y * image_cols;
    row_offset = row * col_count;
    row1_addr = image + ((y - 1) * image_cols);
    row2_addr = image + row_index;
    row3_addr = image + ((y + 1) * image_cols);

    /* B in CCA is a boundary condition for the first column */

    prev_pixel_bucket = UNDEFINED;
 
    flg = 1;
    for (col = 0; col < col_count; col++) {

      /* this code executes for each pixel in the range */

      /*
         compute the 1st spatial derivatives in the
         direction of the axes
         Compute the first derivative in the X axis direction by
         applying a Prewitt operator which applies this mask:
             -1 -1 -1   a b c
              0  0  0 = d e f
              1  1  1   g h i 
         apply the Prewitt operator
      */

      mid_col = minc_int + col;
      sub_col = mid_col - 1;
      add_col = mid_col + 1;

      if ((mid_col >= cc2) && (mid_col <= cc3)) {
        flg = 0;
        delta_x = (*(row3_addr + sub_col)) + /* g */
                  (*(row3_addr + mid_col)) + /* h */
                  (*(row3_addr + add_col)) - /* i */
                  (*(row1_addr + sub_col)) - /* a */
                  (*(row1_addr + mid_col)) - /* b */
                  (*(row1_addr + add_col));  /* c */

        /*
          Compute the first derivative in the Y axis direction by
          applying a Prewitt operator which applies this mask:
            -1  0  1   a d g
            -1  0  1 = b e h 
            -1  0  1   c f i
          apply the Prewitt operator
        */

        delta_y = (*(row1_addr + add_col)) + /* g */
                  (*(row2_addr + add_col)) + /* h */
                  (*(row3_addr + add_col)) - /* i */
                  (*(row1_addr + sub_col)) - /* a */
                  (*(row2_addr + sub_col)) - /* b */
                  (*(row3_addr + sub_col));  /* c */
  
        /* compute the gradient magnitude */

        grad_mag = abs(delta_x) + abs(delta_y);
        }
      else grad_mag = 0.0;

      /*
         Determine if the pixel should be processed, and
         if so, pass it to the connected components algorithm.
      */

      fragment = UNDEFINED;

      if (grad_mag < grad_mag_threshold) /* the pixel should be ignored */
        cur_bucket = UNDEFINED;
      else {
        /* Use the computed delta values to index into
        /* a precomputed arctan table which returns the
        /* bucket for the given values */
       
        /* halve the index values until they validly 
        /* index into the arctan table */

        while ((abs(delta_x) > bucket_table->quadrant_x) ||
               (abs(delta_y) > bucket_table->quadrant_y)) {
          delta_x = delta_x / 2;
          delta_y = delta_y / 2;
          }
        cur_bucket = bucket_table->lookup[
                       delta_y + bucket_table->quadrant_y +
                       ((delta_x + bucket_table->quadrant_x) *
                         bucket_table->size_x)];

        /* If the orientation at the pixel isn't within a desired range,
           it should be ignored. */

        if (cur_bucket != UNDEFINED) {
          /* store the gradient images for Pass 2 */

          *(gradient_image + col + row_offset) = grad_mag / GRADIENT_SCALE;
          *(delta_x_image + col + row_offset) = delta_x;
          *(delta_y_image + col + row_offset) = delta_y;

          /*
            Feed the pixel data to the connected components algorithm (CCA)
            where A is the reference (C is the current. BC pixel).     
            Note that prev_row_bucket is A and prev_pixel_bucket is B
          */

          frag_A = *(FLI + col + ((row - 1) * col_count));
          frag_B = *(FLI + (col - 1) + row_offset);

          if (cur_bucket == prev_row_bucket[col]) {
            /* A and C are in the same bucket, so C's frag_num = A's frag_num */

            fragment = frag_A;
            frag_dest = RFT + frag_A;
            frag_merger = RFT + frag_B;

            /* bump the fragment's pixel count */

            ++(frag_dest->pixel_count);

            /*
               check for a labelling conflict 
               (B.bucket == C.bucket) &&
               (B.frag_num != C.frag_num) &&
               (B and C haven't been merged)
            */

            if ((prev_pixel_bucket == cur_bucket) &&
                (frag_B != frag_A) && (frag_B != UNDEFINED) &&
                ((frag_merger->fet_link == NIL) || 
                 (frag_merger->fet_link != frag_dest->fet_link))) {

              /*
                Create an equivalence relation between A/C's fragment number
                and B's fragment number.

                Merge two region fragments together by making appropriate
                modifications to the RFT and FET to show the fragment
                equivalences.

                assign frag #'s s.t.,  1) frag_A & frag_B are not modified,
                and  2) merge the higher frag # into the lower frag #
              */

              if (frag_A >= frag_B) {
                fragment = frag_B;
                frag_B = frag_A;
                frag_A = fragment;
                frag_dest = RFT + frag_A;
                frag_merger = RFT + frag_B;
                } 

              /* check that the fet_free_ptr is in bounds */ 

              if (fet_free_ptr >= MAX_FET_FREE_PTR_VALUE) {
                fprintf(stderr,"line_finder: Fragment Equivalence Table overflow!\n");    
                return FET_OVERFLOW;
                }

              /* create a fragment equivalence by modifying the RFT and */
              /* putting the equivalence in the FET */ 

              fet_index = frag_dest->fet_link;
              if (fet_index == NIL) {
                if (frag_merger->fet_link == NIL) {
                  /* neither fragment has been merged before */

                  frag_dest->fet_link   = fet_free_ptr;
                  frag_merger->fet_link = fet_free_ptr;

                  FET[fet_free_ptr].frag_num = frag_A;
                  FET[fet_free_ptr].link = fet_free_ptr + 1;
                  fet_free_ptr++;
                  FET[fet_free_ptr].frag_num = frag_B;
                  FET[fet_free_ptr++].link = NIL;
                  }
                else { /* only frag_dest has never been merged before */
                  frag_dest->fet_link = fet_free_ptr;
                  fet_head = fet_free_ptr;
                  FET[fet_free_ptr].frag_num = frag_A; 
                  fet_index = frag_merger->fet_link;
                  FET[fet_free_ptr++].link = fet_index;
                  while (fet_index != NIL) {
                    rft_index = FET[fet_index].frag_num;
                    fet_index = FET[fet_index].link;
                    RFT[rft_index].fet_link = fet_head;
                    }
                  }
                }
              else { /* frag_dest has merged before */ 
                if (frag_merger->fet_link == NIL) {
                  /* frag_merger has not merged before */

                  FET[fet_free_ptr].frag_num = frag_B; 
                  FET[fet_free_ptr].link = NIL; 

                  frag_merger->fet_link = fet_index;

                  while (FET[fet_index].link != NIL)
                    fet_index = FET[fet_index].link;

                  FET[fet_index].link = fet_free_ptr++; 
                  }
                else { /* both frag_dest and frag_merger have merged before */

                  fet_head = fet_index;

                  while (FET[fet_index].link != NIL)
                    fet_index = FET[fet_index].link;

                  FET[fet_index].link = frag_merger->fet_link;

                  fet_index = frag_merger->fet_link;

                  while (fet_index != NIL) {
                    rft_index = FET[fet_index].frag_num;
                    fet_index = FET[fet_index].link;
                    RFT[rft_index].fet_link = fet_head;
                    }
                  }
                }
              }
            }
          else if (cur_bucket == prev_pixel_bucket) {

            /* C's frag_num = B's frag_num */

            fragment = frag_B;
            rft_ptr = RFT + frag_B;
            ++(rft_ptr->pixel_count);
            }
          else { /* C is the start of a new fragment */

            /* allocate and init a new fragment */

            if (++num_frags >= MAX_NUMBER_OF_FRAGMENTS) {
              fprintf(stderr,"line_finder: Exceeded fragment limit!\n");
              return EXCEEDED_FRAG_LIMIT;
              }

            /* C's frag_num = cur. alloc. frag # */

            fragment = num_frags;
            rft_ptr = RFT + fragment;

            rft_ptr->region = NIL;
            rft_ptr->fet_link= NIL;
            rft_ptr->pixel_count = 1; 
            }
          }
        }

      *(FLI + col + row_offset) = fragment;

      /* C becomes B in the next loop */

      prev_pixel_bucket = cur_bucket;  /* C becomes A in the next row loop */
      prev_row_bucket[col] = cur_bucket;
      }
    if (flg && (report > 1)) fprintf(stderr,"line_finder: %9.3f %9.3f %d %d\n",
                                     cc2,cc3,col_count,row);
    }

  free(prev_row_bucket);

  if (report > 1) fprintf(stderr,"Pass 2 - %d fragments\n",num_frags);

  /*
    PASS 2 sequentially scans through the Region Fragment
    Table (RFT).  It does the following:
     -- merges fragments into regions (as determined by the
        Fragment Equivalence Table (FET) ) and forms a region pixel count
     -- eliminates regions whose pixel count is below the filter threshold
     -- Assigns sequential numbers to unfiltered regions (starting from 1)
  */

  num_regions = 0;   /* is incremented prior to use */

  for (fragment = 1; fragment <= num_frags; fragment++) {
    rft_ptr = RFT + fragment;

    if (rft_ptr->region != NIL) continue;

    /* for each unvisited fragment */

    head_fet_index = rft_ptr->fet_link;
    if (head_fet_index == NIL) { /* the fragment is an entire region */

      /* if the number of pixels in the fragment isn't less
      /* than the threshold, allocate the region  */

      if (rft_ptr->pixel_count >= pixel_count_threshold)
        rft_ptr->region = ++num_regions;
      else
        rft_ptr->region = UNDEFINED;
      }
    else { /* the fragment is part of a larger region */

      /* count the number of pixels in the region */

      region_pixel_count = 0;
      fet_index = head_fet_index;

      while (fet_index != NIL) {
        rft_index = FET[fet_index].frag_num;
        region_pixel_count += RFT[rft_index].pixel_count;
        fet_index = FET[fet_index].link;
        }

      /*
        if not below threshold, assign a region # to the
        fragments; otherwise deactivate the fragments.

        if the region pixel count is above the filter
        threshold, assign it a unique region number;
        otherwise indicate that the fragment has no
        region number.
      */

      if (region_pixel_count >= pixel_count_threshold) {

        /* create a new region number */

        num_regions++;

        /* assign the fragments the region # */

        fet_index = head_fet_index;

        while (fet_index != NIL) {
          rft_index = FET[fet_index].frag_num;
          fet_index = FET[fet_index].link;
          RFT[rft_index].region = num_regions;
          }
        }

      else { /* deactivate the fragments */
        fet_index = head_fet_index;
        while (fet_index != NIL) {
          rft_index = FET[fet_index].frag_num;
          fet_index = FET[fet_index].link;
          RFT[rft_index].region = UNDEFINED;
          }
        }
      }
    }

  if (num_regions >= MAX_NUMBER_OF_REGIONS) {
    fprintf(stderr,"line_finder: Number of regions exceeded!\n");
    return 0;
    }

  free(FET);

 if (report > 1) fprintf(stderr,"Pass 3 - %d regions\n",num_regions);

  RST = (RST_ENTRY *) malloc((num_regions + 1) * sizeof(RST_ENTRY));
  if (RST == NULL) {
    fprintf(stderr,"line_finder: Unable to allocate RST\n");
    return 0;
    }

  for (rst_index = 0; rst_index <= num_regions; rst_index++) {
    rst_ptr = RST + rst_index;

    rst_ptr->wsum_x2 = 0;
    rst_ptr->wsum_y2 = 0;
    rst_ptr->wsum_xy = 0;
    rst_ptr->wsum_x  = 0;
    rst_ptr->wsum_y  = 0;
    rst_ptr->sum_x   = 0;
    rst_ptr->sum_y   = 0;
    rst_ptr->sum_delta_x  = 0;
    rst_ptr->sum_delta_y  = 0;
    rst_ptr->sum_grad_mag = 0;
    rst_ptr->max_x = -32000;
    rst_ptr->max_y = -32000;
    rst_ptr->min_x =  32000;
    rst_ptr->min_y =  32000;
    rst_ptr->pixel_count = 0;
    }

 /*
     PASS 3 raster scans through the Fragment Label Image (FLI).
     It does the following:
      -- for each region of interest, accumulates sums
         needed to fit a line to each region
      -- if desired, builds an output Region Label Image (RLI)
  */

  for (row = 0; row < row_count; row++) {

    row_offset = row * col_count;

    /* relocate the origin to the image center */

    rel_row = minr_int + row - row_reloc;

    for (col = 0; col < col_count; col++) {

      /*
        if PASS 1 determined that the pixel is not in a
        region of interest, then ignore the pixel; 
        otherwise, collect statistics (i.e., sums) from 
        each pixel in the region.
      */

      /* if the region is not of interest, ignore the pixel */

      fragment = *(FLI + col + row_offset);

      if (fragment == UNDEFINED) continue;

      rft_ptr = RFT + fragment;

      current_region = rft_ptr->region;

      if (current_region == UNDEFINED) continue;

      /* relocate the origin to the image center */

      rst_ptr = RST + current_region;

      rel_col = minc_int + col - col_reloc;

      /*
         Compute the partial sums needed to fit a line to a region
         and determine its parameters.
      */

      /* update the max and minimum bounding region extent box */

      if (rel_row < rst_ptr->min_x) rst_ptr->min_x = rel_row;  
      if (rel_row > rst_ptr->max_x) rst_ptr->max_x = rel_row; 

      if (rel_col < rst_ptr->min_y) rst_ptr->min_y = rel_col;  
      if (rel_col > rst_ptr->max_y) rst_ptr->max_y = rel_col; 
       
      w_x = rel_row * *(gradient_image + col + row_offset);  
      w_y = rel_col * *(gradient_image + col + row_offset);  

      /* accumulate the partial sums */  

      rst_ptr->sum_x        += rel_row; 
      rst_ptr->sum_y        += rel_col; 
      rst_ptr->wsum_x       += w_x;  
      rst_ptr->wsum_y       += w_y;  
      rst_ptr->wsum_x2      += w_x * rel_row; 
      rst_ptr->wsum_xy      += w_x * rel_col; 
      rst_ptr->wsum_y2      += w_y * rel_col; 
      rst_ptr->sum_delta_x  += *(delta_x_image + col + row_offset); 
      rst_ptr->sum_delta_y  += *(delta_y_image + col + row_offset); 
      rst_ptr->sum_grad_mag += *(gradient_image + col + row_offset); 
      rst_ptr->pixel_count++;

      /* if desired, output a Region Label Image (RLI) */

      if (RLI != NULL) *(RLI + col + row_offset) = current_region;
      }
    }

  free(RFT);
  free(FLI);
  free(gradient_image);
  free(delta_x_image);
  free(delta_y_image);

  if (report > 1) fprintf(stderr,"Pass 4\n");

  /*
     PASS 4 sequentially scans through the Region Statistics
     Table (RST).  It does the following:
      -- for each region, computes the line using the
         accumulated sums and the Eigen technique
      -- puts the fitted line into the output structure
  */

  for (rst_index = 1; rst_index <= num_regions; rst_index++) {

    rst_ptr = RST + rst_index;

    /* compute the line parameters

      Take the partial sums and compute the Eigen scatter matrix 
      in the form [a b]
                  [b c]
    */

    grad_mag = rst_ptr->sum_grad_mag;
    if (grad_mag == 0.0) {
      fprintf(stderr,"line_finder: GM = 0 @ %d\n",rst_index);
      continue;
      }
    sum_x = rst_ptr->wsum_x;
    sum_y = rst_ptr->wsum_y;

    /* E(X^2*w) */

    scat_xx = ((REAL)rst_ptr->wsum_x2) - ((sum_x * sum_x) / grad_mag);

    /* E(X*Y*w) */

    scat_xy = ((REAL)rst_ptr->wsum_xy) - ((sum_x * sum_y) / grad_mag);

    /* E(Y^2*w) */

    scat_yy = ((REAL)rst_ptr->wsum_y2) - ((sum_y * sum_y) / grad_mag);

    /*
      Take an Eigen scatter matrix and compute its two Eigenvectors:
      the small Eigenvector corresponds to the worse line fit and the large 
      Eigenvector is the best fit (positively rotated by 90 degrees from the
      small Eigenvector).
      Method for computing Eigenvalues from a scatter matrix: 
      The 2x2 Eigen scatter matrix is a (Ex^2), b (Exy), and 
      c (Ey^2).  The derivation of the solution of the Eigenvectors 
      for this 2x2 matrix is derived as  
      [a b]   [x] =   [x] ==> ax + by = lx ==>  
      [b c] * [y] = l [y] ==> bx + cy = ly ==>  
         
      x(a-l) + by = 0 ==> [a-l  b ]     ==> 
      y(c-l) + bx = 0 ==> [ b  c-l] = 0 ==> 
         
      (a-l)(c-l) - b^2 = 0 ==> ac-al-cl+l^2+b^2 = 0 ==>    
         
      (1)l^2 + (-a-c)l + (ac-b^2) = 0  
      Now that this is in the quadratic form, the solution is 
      found using the quadratic equation.
    */


    /* compute the parts of the quadatic equation */
    /* calculate -b/2a */

    qe1 = (scat_xx + scat_yy) * 0.5;

    /* calculate sqrt(b^2 - 4ac)/2a */

    qe2 = scat_xx - scat_yy;
    qe2 = sqrt((qe2 * qe2 * 0.25) + (scat_xy * scat_xy));

    /* compute the Eigenvalues for the scatter matrix */

    eivalsmall = qe1 - qe2;
    eivalsmall = (eivalsmall >= 0.0) ? eivalsmall : 0.0;
    eivallarge = qe1 + qe2;
    eivallarge = (eivallarge >= 0.0) ? eivallarge : 0.0;

    /* determine the Eigenvector directions for the scatter matix */

    if (fabs(scat_xy) < 1E-33) {
      /* then line is almost vertical or horizontal
      /* (avoid overflow by processing these special cases */

      if (scat_xx > scat_yy) eivecsmall = PI_OVER_2; /* line is horizontal */
      else                   eivecsmall = 0.0;       /* line is vertical */
      }
    else /* numerically compute Eigenvectors for the scatter matrix */
      eivecsmall = atan2((eivalsmall - scat_xx), scat_xy);

    /*
      Use the eigenvalues to compute the dispersion of the region
      about the extracted line; note that 0 <= dispersion <= 1.
    */

    new_line.dispersion = (eivallarge == 0.0 ? 0.0 : eivalsmall / eivallarge);

    /* filter for dispersion of region about fitted line */
    /* check for low bound on dispersion */

    if ((line_filters->dispersion.min_active == TRUE) &&
        (new_line.dispersion < line_filters->dispersion.min_dispersion))
      continue; /* line not of interest */

    /* check for high bound on dispersion */

    if ((line_filters->dispersion.max_active == TRUE) &&
        (new_line.dispersion >= line_filters->dispersion.max_dispersion))
      continue; /* line not of interest */
 
    /*
      Convert the unsigned eigenvector into a signed theta which 
      encodes the direction of brightness ("right is bright").
      Since eivec points in the (opposite) direction of the gradient,
      its dot product with the average gradient direction indicates whether
      the gradient direction need be rotated by PI.
    */ 

    if (((cos(eivecsmall) * rst_ptr->sum_delta_x) +
         (sin(eivecsmall) * rst_ptr->sum_delta_y)) >= 0.0) 
      {
      new_line.theta = eivecsmall;  /* eivecsmall is OK as is */
      }
    else  /* must rotate eivecsmall by PI to maintain "right is bright" */
      {
      new_line.theta = eivecsmall + PI; 
      }

    MAKE_THETA_IN_RANGE(new_line.theta);

    /* filter for range of acceptable line directions */

    if (line_filters->direction.active == TRUE)
      if (crosses_discontinuity) {
        if ((new_line.theta < line_filters->direction.min_theta) &&
              (new_line.theta >=  line_filters->direction.max_theta))
          continue; /* line not of interest */
        }
      else {
        if ((new_line.theta <  line_filters->direction.min_theta) ||
            (new_line.theta >= line_filters->direction.max_theta))
          continue; /* line not of interest */
        }

    /*
      new_line.theta is the gradient vector pointing
      towards increasing lightness; convert it to a vector
      pointing in the direction of the line (othogonal to it)
    */

    line_direction = new_line.theta + PI_OVER_2;

    MAKE_THETA_IN_RANGE(line_direction);

    /**** Move the origin from the center of the image back to (0,0)  ****/

    num_pixs = rst_ptr->pixel_count;

    centroid_x = (((REAL)rst_ptr->sum_x) / num_pixs) + (REAL)row_reloc;
    centroid_y = (((REAL)rst_ptr->sum_y) / num_pixs) + (REAL)col_reloc;

    /*
       expand the bounding box by a half pixel on all sides to account for
       discrete lattice
    */

    x_min = ((REAL) rst_ptr->min_x) - REGION_EXTANT_EXPANSION + (REAL)row_reloc;
    y_min = ((REAL) rst_ptr->min_y) - REGION_EXTANT_EXPANSION + (REAL)col_reloc;
    x_max = ((REAL) rst_ptr->max_x) + REGION_EXTANT_EXPANSION + (REAL)row_reloc;
    y_max = ((REAL) rst_ptr->max_y) + REGION_EXTANT_EXPANSION + (REAL)col_reloc;

    compute_endpoints(x_min,y_min,x_max,y_max,line_direction,
      centroid_x,centroid_y,
      &new_line.endpoints.l.p1.x, &new_line.endpoints.l.p1.y,
      &new_line.endpoints.l.p2.x, &new_line.endpoints.l.p2.y);

    /* Take two endpoints and compute the Euclidean distance between them. */

    /* use Pythagorean's theorem */

    x_diff = new_line.endpoints.l.p1.x - new_line.endpoints.l.p2.x;
    y_diff = new_line.endpoints.l.p1.y - new_line.endpoints.l.p2.y;
    new_line.length = sqrt((x_diff * x_diff) + (y_diff * y_diff));

    /* filter for line length */
    /* check for low bound on line length */

    if ((line_filters->length.min_active == TRUE) &&
        (new_line.length < line_filters->length.min_length))
      continue; /* line not of interest */

    /* check for high bound on line length */

    if ((line_filters->length.max_active == TRUE) &&
        (new_line.length >= line_filters->length.max_length))
      continue; /* line not of interest */

    /* Compute the gradient magnitude for the line. */

    new_line.grad_mag = grad_mag / num_pixs;

    /* check for low bound on gradient magnitude */

    if ((line_filters->gradient_magnitude.min_active == TRUE) &&
        (new_line.grad_mag < line_filters->gradient_magnitude.min_grad_mag))
      continue; /* line not of interest */

    /* check for high bound on gradient magnitude */

    if ((line_filters->gradient_magnitude.max_active == TRUE) &&
        (new_line.grad_mag >= line_filters->gradient_magnitude.max_grad_mag) )
      continue; /* line not of interest */

    /*** store the region statistics ***/

    new_line.region = rst_index;
    new_line.num_pixels_in_region = rst_ptr->pixel_count;
    new_line.sum_delta_x = rst_ptr->sum_delta_x;
    new_line.sum_delta_y = rst_ptr->sum_delta_y;
    new_line.sum_grad_mag = grad_mag;

    /*** package the line in the out line list ***/

    /* Enter a line into the outgoing line list. */

    if (num_lines >= number_lines) {
      fprintf(stderr,"line_finder: Line list overflow: too many lines for list size\n");
      return LINE_LIST_OVERFLOW;
      }
    out_line_list[num_lines++] = new_line;
    if (display)
      draw_line_rgb(display - 1,
                    new_line.endpoints.l.p1.x,new_line.endpoints.l.p1.y,
                    new_line.endpoints.l.p2.x,new_line.endpoints.l.p2.y,GREEN);
    }

  if (display)
    for (i = 0; i < 6; i++) {
      j = i + 1;
      if (j >= 6) j = 0;
      erase_line(display - 1,pt[i].y,pt[i].x,pt[j].y,pt[j].x);
      }

  if (report > 1) fprintf(stderr,"Returning success\n");

  free(RST);

 return num_lines;
}

#define swap(a,b,t) {t = a; a = b; b = t;}

void call_line_finder(pixel_count_threshold,gradient_mag_threshold)
int pixel_count_threshold;
REAL gradient_mag_threshold;
{
int i, j, k, m;
REAL theta, theta_range;
REAL x_range, y_range;
BUCKET_TABLE *bucket_table;
ML *lptr;
LINE2D l;
LINE3D *wl;
REAL delta_x, delta_y;
LINE local_lines[200];
DL *data_lines;
char work[50];
LINE_FILTER left_line_filters;
BUCKET bucket_num;
int x, y, size_x, size_y, x_size;
int index, cnt;
REAL bucket_theta;
REAL min[2],max[2];
REAL t;

  my_id = m_get_myid();

  bucket_table = (BUCKET_TABLE *) alloc_bt();
  if (bucket_table == NULL) {
    return_error("Unable to allocate bucket table!");
    return;
    }

  left_line_filters.gradient_magnitude.min_active = TRUE;
  left_line_filters.gradient_magnitude.min_grad_mag = 5.0;
  left_line_filters.gradient_magnitude.max_active = FALSE;
  left_line_filters.gradient_magnitude.max_grad_mag = 99999.0;

  left_line_filters.length.min_active = FALSE;
  left_line_filters.length.min_length = 0.0;
  left_line_filters.length.max_active = FALSE;
  left_line_filters.length.max_length = 99999.0;

  left_line_filters.dispersion.min_active  = FALSE;
  left_line_filters.dispersion.min_dispersion = 0.0;
  left_line_filters.dispersion.max_active  = FALSE;
  left_line_filters.dispersion.max_dispersion = 99999.0;

  m = 0;

  while (m < number_models) {
    S_LOCK(&model_lock);

    if (model_info[m].current_model_line >= model_info[m].number_model_lines) {
      S_UNLOCK(&model_lock);
      m++;
      continue;
      }

    k = model_info[m].current_model_line;
    model_info[m].current_model_line++;
#ifdef SEQUENT
    S_UNLOCK(&model_lock);

    sprintf(work,"M%dL%d\n",m,k);
    show_msg_processor(work);
#endif
    x_range = model_info[m].x_range;
    y_range = model_info[m].y_range;
    theta_range = model_info[m].theta_range;

    lptr = model_info[m].lines2d + k;
    wl = model_info[m].lines3d + k;
    l = lptr->l;

    theta = atan2(l.p1.x - l.p2.x,l.p2.y - l.p1.y);

    /* Build bucket table from expected orientation of line. */

    min[0] = theta - theta_range;
    max[0] = theta + theta_range;
    min[1] = min[0] + PI;
    max[1] = max[0] + PI;

    if      (min[0] >  PI) min[0] -= PI2;
    else if (min[0] < -PI) min[0] += PI2;
    if      (max[0] >  PI) max[0] -= PI2;
    else if (max[0] < -PI) max[0] += PI2;
    if      (min[1] >  PI) min[1] -= PI2;
    else if (min[1] < -PI) min[1] += PI2;
    if      (max[1] >  PI) max[1] -= PI2;
    else if (max[1] < -PI) max[1] += PI2;

/*
    if (report > 1)
      fprintf(stderr,"theta = %6.2f, ranges = %6.2f %6.2f & %6.2f %6.2f\n",
        theta * 180.0 / PI, min[0] * 180.0 / PI,max[0] * 180.0 / PI,
                            min[1] * 180.0 / PI,max[1] * 180.0 / PI);
*/

    left_line_filters.direction.active = TRUE;
    left_line_filters.direction.min_theta = min[0];
    left_line_filters.direction.max_theta = max[0];

    /* Use contrast direction to find lines */

    if      (wl->contrast > 0.0) min[0] = max[0] = 0.0;
    else if (wl->contrast < 0.0) min[1] = max[1] = 0.0;

    size_x = bucket_table->quadrant_x;
    size_y = bucket_table->quadrant_y;
    x_size = bucket_table->size_x;

    for (x = -size_x; x <= size_x; x++) {
      index = size_y + ((x + size_x) * x_size);
      for (y = -size_y; y <= size_y; y++) {

        /* Get the radian angle for the lookup table position */

        bucket_theta = *(thetas + y + index);

        /* see if a bucket has been defined for the computed theta */

        if      ((min[0] > max[0]) &&
                 ((bucket_theta >= min[0]) || (bucket_theta < max[0])))
          bucket_num = 2;
        else if  ((bucket_theta >= min[0]) && (bucket_theta < max[0]))
          bucket_num = 2;
        else if ((min[1] > max[1]) &&
                 ((bucket_theta >= min[1]) || (bucket_theta < max[1])))
          bucket_num = 1;
        else if  ((bucket_theta >= min[1]) && (bucket_theta < max[1]))
          bucket_num = 1;
        else
          bucket_num = UNDEFINED;

        /* store into the bucket lookup table */
        /* set the bucket lookup table location */

        bucket_table->lookup[index + y] = bucket_num;
        }
      }

    j = line_finder(image,row_size,col_size,current_plane + 1,
                    l.p1.y, l.p1.x, l.p2.y, l.p2.x, y_range, x_range,
                    bucket_table,
                    local_lines, 100,
                    NULL, &left_line_filters,
                    pixel_count_threshold,gradient_mag_threshold);

    lptr->number_data_lines = 0;
    lptr->data_lines = NULL;

    if (j <= 0) continue;

    data_lines = (DL *) malloc(j * sizeof(DL));
    if (data_lines == NULL) continue;

    lptr->data_lines = data_lines;

    cnt = 0;
    for (i = 0; i < j; i++) {
      data_lines[cnt].l = local_lines[i].endpoints.l;
      delta_x = data_lines[cnt].l.p1.x - data_lines[cnt].l.p2.x;
      delta_y = data_lines[cnt].l.p1.y - data_lines[cnt].l.p2.y;
      data_lines[cnt].len = sqrt(delta_x * delta_x + delta_y * delta_y);
      cnt++;
      }
    lptr->number_data_lines = cnt;
    S_LOCK(&model_lock);
    model_info[m].number_data_lines += cnt;
    S_UNLOCK(&model_lock);
    }
  free(bucket_table);
}

