/* -*-c-mode-*- /*
/*------------------------------------------------------
 *  ROBOT_OPERATIONS.C - Control of robot servicing on the Sequent from LLVS.
 *  James H. Burrill Created 1990/4/26
 *  Based on work by Robert Heller, Kaj Telenar, Mark Gryska
 *  Last mod - 
 *--------------------------------------------------------

           Copyright 1990 University of Massachusetts.
                         All rights reserved.

     Permission to copy and modify this software and its documen-
     tation only for internal use in your organization is hereby
     granted, provided that this notice is retained thereon and
     on all copies.  UMASS makes no representations as to the sui-
     tability and operability of this software for any purpose.
     It is provided "as is" without express or implied warranty.

     UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
     NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
     INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
     SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
     IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
     ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
     FORMANCE OF THIS SOFTWARE.

     No other rights, including, for example, the right to redis-
     tribute this software and its documentation or the right to
     prepare derivative works, are granted unless specifically
     provided in a separate license agreement.

     Copyright 1990, University of Massachusetts. All rights
     reserved.



 *------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include "sequent.h"
#include "tableau.h"

#define SUCCESS  1
#define FAIL    -1
#define swap(a,b,t) {t = a; a = b; b = t;}

typedef struct {
  int valid;  /* Valid result if not zero! */
  double error_value;
  VEC3D trans;
  VEC3D rot;
  } C_POSE;

static C_POSE final_pose;

typedef struct {
  double correlation;
  int match_row;
  int match_col;
  } TEMPLATE_RESULT;

static TEMPLATE_RESULT template_result;

char *malloc();
double atof();
IMAGE *read_image();
void call_line_finder();
void call_matcher_setup();
void local_search();
double atof();
float template_match();
void construct_template();

/*
  display Integer indicating the degree to which the search is
          to be displayed in the search_display window.  Zero
          will suppress ALL graphics.
*/
int display = 2;
/*
  report An integer indicating the degree of  reporting desired.
         Text will be printed to stream *standard_output*.  Zero 
         will suppress ALL text.  Successivley higher integers
         will produce increasingly copious amounts of text.
*/
int report = 1;
/*
  xmonitor Integer indicating the degree to which the search path
           is to be monitored in the monitor_display window.
           Zero suppresses ALL monitoring.   One will monitor
           without pause points.  Two will pause after each step
           in the search.
*/
int xmonitor = 0;

FILE *lp = NULL;
FILE *ip = NULL;
static char lpname[256];
static char planefilename[64];

static unsigned char *template = NULL;
static int template_rows;
static int template_cols;

       IMAGE *image = NULL;
       MODEL_INFO *model_info = NULL;

       int row_size;
       int col_size;
       int col_center;
       int row_center;
       int current_plane = 0;
       int my_id = 0;
       int number_models;
static int model_index;
       int model_number;
static int total_model_lines;
static int max_displacement;
static int pixel_count_threshold;
static int trials;
static int match_method;
static int e_om_level;

static REAL  attenuation;
static REAL  scale_range;
static REAL  tau = 0.000;
static REAL  loading;
static REAL  gradient_mag_threshold;
static REAL  contrast_threshold;
       REAL  fovx;
       REAL  fovy;
       REAL  trans_err;
       REAL  rot_err;
       REAL  best_errvalue;
       VEC3D original_trans;
       QUAT  original_rotqu;
       VEC3D best_trans;
       QUAT  best_rotqu;

static char read_buff[1024];
static int read_index = 2000;
static int read_buf_size = 0;

int robot_reserve_memory(n)
{
char *ptr;

  ptr = (char *) malloc(n);
  if (ptr != NULL) free(ptr);
  return (ptr != NULL);
}

void show_msg_processor(msg)
char *msg;
{
  if (lp != NULL) fputs(msg,lp);

  fputs(msg,stderr);
}

void show_msg_model(msg)
char *msg;
{
  if (lp != NULL) fprintf(lp,"M%d%s",model_number,msg);

  fprintf(stderr,"robot_ops: M%d%s",model_number,msg);
}

/*
  GET_LINE
    This is a re-write of fgets that does not do some timing, testing,
  checking that hangs when stdin is a network file.
*/

static int get_line(buffer,maxsize)
char *buffer;
int maxsize;
{
int n;

  if (ip != NULL) {
    fgets(buffer,maxsize,ip);
    return strlen(buffer);
    }

  n = 0;

  while (n < maxsize) {
    if (read_index >= read_buf_size) {
      do {
        read_buf_size = read(0,read_buff,1024);
        } while (read_buf_size <= 0);
      read_index = 0;
      }
    if (read_buff[read_index] == '\0') {
      read_index++;
      if (n == 0) continue;
      break;
      }
    *buffer++ = read_buff[read_index++];
    n++;
    }
  *buffer++ = '\n';
  *buffer++ = '\0';
  return n;
}

static int get_data(buffer,bufsize)
char *buffer;
int bufsize;
{
int n;

  if (ip != NULL) {
    fgets(buffer,bufsize,ip);
    return strlen(buffer);
    }

  n = 0;

  while (n < bufsize) {
    if (read_index >= read_buf_size) {
      do {
        read_buf_size = read(0,read_buff,1024);
        } while (read_buf_size <= 0);
      read_index = 0;
      }
    *buffer++ = read_buff[read_index++];
    n++;
    }

  return n;
}

static int memcmp(a,b,c)
char *a;
char *b;
int c;
{
int i;

  for (i = 0; i < c; i++)
    if (*a++ != *b++) return 1;
  return 0;
}

void return_error(msg)
char *msg;
{
char buf[200];

  sprintf(buf,"*ERROR* %s\n",msg);

  if (lp != NULL) fputs(buf,lp);
  fputs(buf,stderr);
  fflush(stderr);
}

void return_msg(msg)
char *msg;
{
  if (lp != NULL) fputs(msg,lp);
  fputs(msg,stderr);
  fflush(stderr);
}

int set_display_parameters(Display,Report,Monitor)
int Display;
int Report;
int Monitor;
{
  display = Display;
  report = Report;
  xmonitor = Monitor;
}

int show_image(im,name)
int im;
char *name;
{
char buf[200];

  if (im < 0 || im > 8) {
    fprintf(stderr,"robot_ops: Incorrect image number!\n");
    return;
    }
  sprintf(buf,"*SHOW %d\n",im);
  lputs(buf);

  current_plane = im;

  display_image(current_plane,image,name,row_size,col_size,1,0.0);

  return SUCCESS;
}

int receive_image(image_data,rs,cs)
unsigned char *image_data;
int rs;
int cs;
{
int i;
char buf[200];
int bytes;
IMAGE *ptr;

  row_size = rs;
  col_size = cs;

  bytes = row_size * col_size;

  if (image != NULL) free(image);
  image = (unsigned char *) malloc(bytes);
  if (image == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate image!\n");
    return FAIL;
    }
  ptr = image;

  sprintf(buf,"*RECI %d %d\n",row_size, col_size);
  lputplane();

  for (i = 0; i < bytes; i++) *ptr++ = *image_data++;

  return SUCCESS;
}

int construct_template_op(center_row, center_col,
                       template_rs, template_cs)
int center_row, center_col, template_rs, template_cs;
{
char buf[200];
int r, c;

  template_rows = template_rs;
  template_cols = template_cs;

  sprintf(buf,"*GETT %d %d %d %d\n",
    center_row, center_col, template_rows, template_cols);
  lputs(buf);

  if (template != NULL) free(template);
  template = (unsigned char *) malloc(template_rows * template_cols);
  if (template == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate template!\n");
    return FAIL;
    }

  if (image == NULL) {
    fprintf(stderr,"robot_ops: No image in use!\n");
    return FAIL;
    }

  construct_template(image,row_size, col_size,
                     template,template_rows,template_cols,
                     center_row, center_col);


  return SUCCESS;
}

int send_template(template_data,template_rs,template_cs)
unsigned char *template_data;
int template_rs;
int template_cs;
{
int i, bytes;
char buf[200];
unsigned char *ptr;

  template_rows = template_rs;
  template_cols = template_cs;

  sprintf(buf,"*RECT %d %d\n",
    template_rows,template_cols);
  lputs(buf);

  bytes = template_rows * template_cols;

  if (template != NULL) free(template);
  template = (unsigned char *) malloc(bytes);
  if (template == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate template!\n");
    return FAIL;
    }
  ptr = template;

  for (i = 0; i < bytes; i++) *ptr++ = *template_data++;

  

  return SUCCESS;
}

/*  function to print templates */

static print_template(template_data,temp_rows,temp_cols)
unsigned char *template_data;
int temp_rows, temp_cols;
{
    register int col, row;
    register unsigned char *p;

    printf("\n    ");
    p = template_data;
    for (col = 0; col < temp_cols; col++) printf("%3d ", col);
    for (row = 0; row < temp_rows; row ++) {
	printf("\n%2d: ",row);
	for (col = 0; col < temp_cols; col++) {
	    printf("%3d ",*p++);
	    }
	}
    printf("\n");
    }



TEMPLATE_RESULT * template_match_op(center_row, center_col,
                                    search_rows, search_cols)
int center_row;
int center_col;
int search_rows;
int search_cols;
{
char buf[200];
int r, c;
char *ptr;
int match_row, match_col;
float correlation;

  sprintf(buf,"*MATT %d %d %d %d\n",
    center_row, center_col, search_rows, search_cols);
  lputs(buf);

  /* print_template(template,template_rows,template_cols); */

  template_result.correlation = template_match(template,
                        template_rows,template_cols,
                        image,row_size, col_size,
                        center_row, center_col,
                        search_rows, search_cols,
                        &template_result.match_row, &template_result.match_col);

  return &template_result;
}

static void free_tableau(tab)
TABLEAU *tab;
{
  if (tab->pairs    != NULL) free(tab->pairs);
  tab->pairs = NULL;
  if (tab->curr_mls != NULL) free(tab->curr_mls);
  tab->curr_mls = NULL;
  if (tab->init_mls != NULL) free(tab->init_mls);
  tab->init_mls = NULL;
  if (tab->mlpairs  != NULL) free(tab->mlpairs);
  tab->mlpairs = NULL;
}

static void free_model(model)
MODEL_INFO *model;
{
  if (model->lines2d      != NULL) free(model->lines2d);
  model->lines2d = NULL;
  if (model->lines3d      != NULL) free(model->lines3d);
  model->lines3d = NULL;

  if (model->tab == model->best_tableau) model->best_tableau = NULL;

  if (model->tab          != NULL) free_tableau(model->tab);
  model->tab = NULL;
  if (model->best_tableau != NULL) free_tableau(model->best_tableau);
  model->best_tableau != NULL;
}

static void free_models()
{
int i;

  for (i = 0; i < model_index; i++) free_model(model_info + i);
  free(model_info);
  model_info = NULL;
}

int match_model_init(pct,gmt,mm,eol,atten,max_disp,scale_r,ldg,
                     ct,m_trials,num_models,fx,fy,terr,rerr,trans,rot)
int pct;
double gmt;
int mm;
int eol;
double atten;
int max_disp;
double scale_r;
double ldg;
double ct;
int m_trials;
int num_models;
double fx, fy;
double terr, rerr;
float trans[3];
float rot[3];
{
char buf[200];
VEC3D rotation;
VEC3D rtrans;

  pixel_count_threshold = pct;
  gradient_mag_threshold = gmt;
  match_method = mm;
  e_om_level = eol;
  attenuation = atten;
  max_displacement = max_disp;
  scale_range = scale_r;
  loading = ldg;
  contrast_threshold = ct;
  trials = m_trials;
  number_models = num_models;

  fovx = fx;
  fovy = fy;
  rtrans.x = trans[0];
  rtrans.y = trans[1];
  rtrans.z = trans[2];
  rotation.x = rot[0];
  rotation.y = rot[1];
  rotation.z = rot[2];

  final_pose.valid = 0;
  final_pose.error_value = 1.0e10;

  final_pose.trans.x = trans[0];
  final_pose.trans.y = trans[1];
  final_pose.trans.z = trans[2];

  final_pose.rot.x = rot[0];
  final_pose.rot.y = rot[1];
  final_pose.rot.z = rot[2];

  trans_err = terr;
  rot_err = rerr;

  robot_to_camera(&rtrans,&rotation,&original_trans,original_rotqu);

  sprintf(buf,
    "*MATM -a %f -b %d -c %f -g %f -l %f -m %d -o %d -p %d -s %f -t %d\n",
    attenuation, max_displacement, contrast_threshold, gmt, loading,
    match_method, e_om_level, pct, scale_range, trials);

  lputs(buf);

  sprintf(buf,"M %d %f %f %f %f %f %f %f %f %f %f\n",
    num_models,fovx,fovy,terr,rerr,
    trans[0],trans[1],trans[2], 
    rot[0],rot[1],rot[2]);

  lputs(buf);

  if (model_info != NULL) free_models();
  model_info = (MODEL_INFO *) malloc(number_models * sizeof(MODEL_INFO));
  if (model_info == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate model info!\n");
    return FAIL;
    }

  model_index = 0;

  return SUCCESS;
}

int send_model(num_model_lines,nlines3d,nlines2d,err_x,err_y,err_r)
int num_model_lines;
float *nlines3d;
float *nlines2d;
double err_x;
double err_y;
double err_r;
{
char buf[200];
int i, j, k, n;
LINE3D *lines3d;
LINE3D *mptr;
LINE2D line2d;
LINE3D line3d;
LINE2D dl;
ML *lines2d;
ML *ml;
ML *ml2;
DL *data_lines;
DL local_lines[200];
char buffer[200];
char ctype;
double xdif, ydif, dist, t, contrast;
int cnt, flg;
MODEL_INFO *model;

  sprintf(buf,"N %f %f %f %d\n",err_x,err_y,err_r,num_model_lines);
  lputs(buf);

  model = model_info + model_index;

  model->x_range = err_x;
  model->y_range = err_y;
  model->theta_range = err_r;

  model->current_model_line = 0;
  model->number_data_lines = 0;
  model->current_trial = 0;
  model->best_e_match = 999999.9;
  model->lines2d = NULL;
  model->lines3d = NULL;
  model->tab = NULL;
  model->best_tableau = NULL;

  lines2d = (ML *) malloc(num_model_lines * sizeof(ML));
  if (lines2d == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate 2d lines!\n");
    goto err_model_return;
    }

  lines3d = (LINE3D *) malloc(num_model_lines * sizeof(LINE3D));
  if (lines3d == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate 3d lines!\n");
    free(lines2d);
    goto err_model_return;
    }

  model->lines2d = lines2d;
  model->lines3d = lines3d;

  ml = lines2d;
  cnt = 0;

  for (i = 0; i < num_model_lines; i++) {
    ml->number_data_lines = 0;
    ml->data_lines = NULL;
    ml++;
    }

  ml = lines2d;
  mptr = lines3d;

  for (i = 0; i < num_model_lines; i++) {
    line3d.p1.x = *(nlines3d + 0);
    line3d.p1.y = *(nlines3d + 1);
    line3d.p1.z = *(nlines3d + 2);
    line3d.p2.x = *(nlines3d + 3);
    line3d.p2.y = *(nlines3d + 4);
    line3d.p2.z = *(nlines3d + 5);
    contrast    = *(nlines3d + 6);

    nlines3d += 7;

    line2d.p1.x = *(nlines2d + 0);
    line2d.p1.y = *(nlines2d + 1);
    line2d.p2.x = *(nlines2d + 2);
    line2d.p2.y = *(nlines2d + 3);

    nlines2d += 4;

    sprintf(buf,"W %f %f %f %f %f %f %f\n",
      line3d.p1.x,line3d.p1.y,line3d.p1.z,
      line3d.p2.x,line3d.p2.y,line3d.p2.z,
      contrast);

    lputs(buf);
    sprintf(buf,"R %f %f %f %f\n",
            line2d.p1.x,line2d.p1.y,line2d.p2.x,line2d.p2.y);
    lputs(buf);

    mptr->contrast = contrast;

    xdif = line2d.p2.x - line2d.p1.x;
    ydif = line2d.p2.y - line2d.p1.y;
    dist = sqrt(xdif * xdif + ydif * ydif);

    if (((line2d.p1.x < 0.0) || (line2d.p1.x > col_size)) &&
        ((line2d.p2.x < 0.0) || (line2d.p2.x > col_size)) &&
        ((line2d.p1.y < 0.0) || (line2d.p1.y > row_size)) &&
        ((line2d.p2.y < 0.0) || (line2d.p2.y > row_size))) {
      if (report > 1)
        fprintf(stderr,"robot_ops: line endpoints %8.2f %8.2f %8.2f %8.2f\n",
                        line2d.p1.x,line2d.p1.y,line2d.p2.x,line2d.p2.y);
      continue;
      }
    if ((dist < 7.0) || (dist > 512)) {
      if (report > 1)
        fprintf(stderr,"robot_ops: line length     %8.2f %8.2f %8.2f %8.2f\n",
                        line2d.p1.x,line2d.p1.y,line2d.p2.x,line2d.p2.y);
      continue;
      }
    if (fabs(contrast) < contrast_threshold)  {
      if (report > 1)
        fprintf(stderr,"robot_ops: line contrast   %8.2f %8.2f %8.2f %8.2f\n",
                        line2d.p1.x,line2d.p1.y,line2d.p2.x,line2d.p2.y);
      continue;
      }

    flg = 0;
    if (line2d.p2.x < line2d.p1.x) flg = 1;
    else if ((line2d.p2.x == line2d.p1.x) && (line2d.p2.y < line2d.p1.y)) flg = 1;

    if (flg) {
      swap(line2d.p1.x,line2d.p2.x,t)
      swap(line2d.p1.y,line2d.p2.y,t)
      swap(line3d.p1.x,line3d.p2.x,t)
      swap(line3d.p1.y,line3d.p2.y,t)
      swap(line3d.p1.z,line3d.p2.z,t)
      line3d.contrast = - line3d.contrast;
      }

    for (k = 0; k < cnt; k++) {
      ml2 = lines2d + k;
      if (ml2->l.p1.x != line2d.p1.x) continue;
      if (ml2->l.p1.y != line2d.p1.y) continue;
      if (ml2->l.p2.x != line2d.p2.x) continue;
      if (ml2->l.p2.y != line2d.p2.y) continue;

      break;
      }

    if (k != cnt) {
      if (report > 1)
        fprintf(stderr,"robot_ops: line duplicate  %8.2f %8.2f %8.2f %8.2f\n",
                        line2d.p1.x,line2d.p1.y,line2d.p2.x,line2d.p2.y);
      continue; /* Line is not unique */
      }

    /* New line is unique */

    ml->l = line2d;

    *mptr = line3d;

    cnt++;
    ml++;
    mptr++;
    }

  if (report && (cnt != num_model_lines))
    fprintf(stderr,"robot_ops: M%d - Deleted %d lines from %d.\n",
            model_index,num_model_lines - cnt,num_model_lines);

  model->number_model_lines = cnt;
  total_model_lines += cnt;
  model_index++;

  return SUCCESS;

err_model_return:
  model->number_model_lines = cnt;

  return FAIL;
}

C_POSE * end_match()
{
char buf[200];
char *ptr;
double x;
int r, c;
int i, j, k, m;
int use_data, total_lines;
int ml_weighting;
PL *lines2d;
PL *pose_line;
ML *ml;
DL *dptr;
LINE3D *mptr;
char work[200];
TABLEAU *best_tableau;
PR *pair;

  ml_weighting = FALSE;
  use_data = TRUE;

  if (number_models <= 0) return &final_pose;

  if (image == NULL) {
    fprintf(stderr,"robot_ops: No image in use!\n");
    return &final_pose;
    }

  setup_for_buckets(32);

  if (display > 0) 
    for (m = 0; m < number_models; m++) {
      ml = model_info[m].lines2d;

      for (i = 0; i < model_info[m].number_model_lines; i++) {
        draw_ml(ml);
        ml++;
        }
      }

  row_center = row_size >> 1;
  col_center = col_size >> 1;

  initialize_matcher(max_displacement,attenuation,scale_range);

  call_line_finder(pixel_count_threshold,gradient_mag_threshold);

  /* Write data to log file */

  for (m = 0; m < number_models; m++) {
    ml = model_info[m].lines2d;
    mptr = model_info[m].lines3d;

    for (i = 0; i < model_info[m].number_model_lines; i++) {
      if (display > 0) undraw_ml(ml);

      ml++;
      mptr++;
      }
    }

  call_matcher_setup(tau,ml_weighting);

  local_search(match_method,trials,loading,e_om_level);

  if (use_data) {
    total_lines = 0;
    for (m = 0; m < number_models; m++)
      total_lines += model_info[m].number_data_lines;
    }
  else total_lines = total_model_lines;

  lines2d = (PL *) malloc(total_lines * sizeof(PL));

  if (lines2d == NULL) {
    fprintf(stderr,"robot_ops: Unable to allocate final storage!\n");
    goto match_exit;
    }

  k = 0;
  pose_line = lines2d;

  for (m = 0; m < number_models; m++) {

    best_tableau = model_info[m].best_tableau;

    for (i = 0; i < model_info[m].number_model_lines; i++) {

      ml = best_tableau->curr_mls + i;
      mptr = model_info[m].lines3d + i;

      if (display > 0) draw_ml(ml);

      for (j = ml->pri_start; j <= ml->pri_end; j++) {
        pair = best_tableau->pairs + j;
        dptr = pair->data;

        if (display > 0) draw_pair(pair);

        if (!pair->member) continue;

        if (use_data) {
          pose_line->l.p1.x = row_center - dptr->l.p1.y;
          pose_line->l.p1.y = col_center - dptr->l.p1.x;
          pose_line->l.p2.x = row_center - dptr->l.p2.y;
          pose_line->l.p2.y = col_center - dptr->l.p2.x;
          pose_line->weight = 1.0; /* pair->percentl * ml->len; */

          pose_line->active = pair->member;
          pose_line->line3d = mptr;
          pose_line++;

          k++;
          }
        }
      if (!use_data) {
        pose_line->l.p1.x = row_center - ml->l.p1.y;
        pose_line->l.p1.y = col_center - ml->l.p1.x;
        pose_line->l.p2.x = row_center - ml->l.p2.y;
        pose_line->l.p2.y = col_center - ml->l.p2.x;
        pose_line->weight = 1.0;
        pose_line->active = TRUE;
        pose_line->line3d = mptr;
        pose_line++;

        k++;
        }
      }
    }

  call_pose_lines(lines2d,k,500);

  camera_to_robot(&best_trans,best_rotqu,&final_pose.trans,&final_pose.rot);
  final_pose.error_value = best_errvalue;

  /* Display result & free up non-essentials */

  for (m = 0; m < number_models; m++) {
    best_tableau = model_info[m].best_tableau;
    for (i = 0; i < best_tableau->m; i++) {
      ml = best_tableau->curr_mls + i;
      if (ml->data_lines != NULL) free(ml->data_lines);
      ml->data_lines = NULL;
      }
    if (model_info[m].lines2d != NULL) free(model_info[m].lines2d);
    model_info[m].lines2d = NULL;
    if (model_info[m].lines3d != NULL) free(model_info[m].lines3d);
    model_info[m].lines3d = NULL;
    }

match_exit:
  if (lines2d != NULL) free(lines2d);
  model_info = NULL;

  final_pose.valid = 1;
  return &final_pose;
}

/* =========================================== */
/*             Log file hackery                */

void start_logging(lognam,plnam)
register char *lognam, *plnam;
{
  strcpy(lpname,lognam);
  strcpy(planefilename,plnam);
  if (lp != NULL) fclose(lp);
  lp = fopen(lpname,"w");
}

void end_logging()
{
  if (lp == NULL) return;
  fclose(lp);
  lp = NULL;
}

static lputs(str)
register char *str;
{
  if (lp == NULL) return;
  fputs(str,lp);
}

static lputplane()
{
  if (lp == NULL) return;
  fprintf(lp,"*FILI %s\n",planefilename);
}

static lputl(str,slen)
register char *str;
register int slen;
{
  if (lp == NULL) return;
  while (slen-- > 0) {
    fputc(lp,*str);
    str++;
    }
}
