/*

           Copyright 1990 University of Massachusetts.
                         All rights reserved.

     Permission to copy and modify this software and its documen-
     tation only for internal use in your organization is hereby
     granted, provided that this notice is retained thereon and
     on all copies.  UMASS makes no representations as to the sui-
     tability and operability of this software for any purpose.
     It is provided "as is" without express or implied warranty.

     UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
     NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
     INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
     SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
     IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
     ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
     FORMANCE OF THIS SOFTWARE.

     No other rights, including, for example, the right to redis-
     tribute this software and its documentation or the right to
     prepare derivative works, are granted unless specifically
     provided in a separate license agreement.

     Copyright 1990, University of Massachusetts. All rights
     reserved.



 */
#ifdef SEQUENT
#include "vg423.h"
#endif
/*
  
  LOCAL SEARCH MODEL MATCHING TABLEAU
  
  A Tableau is a table of model-data lines segment pairs.   Each column
correspondes with a distinct model-data pair and the current match state
(member t or nil) is recorded with each pair.  The point of the structure
is to create one uniform and efficient data structre which various local search
algorithms may use.
  
  The basic componenents in the table are:

    ml      Modle line segments,  both the initial or canoncial segments and those
            those transformed in accordance with the current correspondence.

    dl      Data line segments.

    pr      A pair, one model segment and one data segment.  When the pairs
            member field is t the pair is part of the match.
  
    strans  A structure for recording the current model transformation.  This is 
            in fact a scale, rotation and translation between the model in it's 
            initial or canonical position and the position which minimizes the 
            spatial fit error.
  
    tableau In essenence an array of model-data pairs along with a transformation
            and additional distinquishing features.  One note,  a second array of 
            pointers to the model-data pairs is maintained.  This array is sorted 
            over the extent associated with each distinct model line in order
            to speed computation of the omission error.
*/

typedef struct {
  REAL x;
  REAL y;
  } VEC2D;

typedef struct {
  REAL x;
  REAL y;
  REAL z;
  } VEC3D;

typedef REAL QUAT[4];
typedef REAL MAT3X3[3][3];

typedef struct {
  VEC3D p1;
  VEC3D p2;
  float contrast;
  } LINE3D;

typedef struct {
  VEC2D p1;
  VEC2D p2;
  } LINE2D;

typedef struct dl {
  LINE2D l;
  REAL len;
  REAL umin;
  REAL umax;
  } DL;

typedef struct pl {
  LINE3D *line3d;
  LINE2D l;
  REAL weight;
  unsigned char active;
  } PL;

typedef struct ml {
  LINE2D l;
  REAL pmx;
  REAL pmy;
  REAL len;
  REAL nx;
  REAL ny;
  REAL rho;
  REAL umin;
  REAL umax;
  REAL pom;
  REAL e_fit;
  REAL e_om;
  REAL weight;
  int pri_start;
  int pri_end;
  int number_data_lines;
  DL *data_lines;
  } ML;

typedef struct strans {
  double s;
  double tx;
  double ty;
  double phix;
  double phiy;
  } STRANS;

/*
  The spatial fit error is expressed in terms of the following constants
  associated with each model data Pair.

      A  Matrix         2x2 symetric
      B  Matrix         2x2 not symetric
      C  Matrix         2x2 symetric
      U  Vector         2x1 
      V  Vector         2x1
      k  Scalar        
*/

typedef struct {
  double a11;
  double a12;
  double a22;
  double b11;
  double b12;
  double b21;
  double b22;
  double c11;
  double c12;
  double c22;
  double u1;
  double u2;
  double v1;
  double v2;
  double k;
  } ABCUVK;

typedef struct pr {
  int member;
  int modi;
  ML *init;
  ML *curr;
  DL *data;
  float percentl;
  ABCUVK cor;
  } PR;

/* Multiply two 2x2 matrices.  The result is stored in matrix r. */
#define M2X2_MULT(x11,x12,x21,x22,y11,y12,y21,y22,r11,r12,r21,r22) \
 {r11 = ((x11) * (y11)) + ((x12) * (y21));\
  r12 = ((x11) * (y12)) + ((x12) * (y22));\
  r21 = ((x21) * (y11)) + ((x22) * (y21));\
  r22 = ((x21) * (y12)) + ((x22) * (y22));}

/* Compute the inverse of a 2x2 symetric matrix, return an error
  if the matrix is singular.  
*/
#define M2X2_INV_SYM(x11,x12,x22,r11,r12,r22,singularp) \
{double det = ((x11) * (x22)) - ((x12) * (x12)); \
 if (fabs(det) < 1.0e-10) singularp = TRUE; \
 else {singularp = FALSE;\
 r11 = (x22) / det;\
 r22 = (x11) / det;\
 r12 = - (x12) / det;}} 

#define M2X2_SUBTRACT(x11,x12,x21,x22,y11,y12,y21,y22,r11,r12,r21,r22) \
{r11 = (x11) - (y11);\
 r12 = (x12) - (y12);\
 r21 = (x21) - (y21);\
 r22 = (x22) - (y22);}

#define M2X2_SUB_SYM(x11,x12,x22,y11,y12,y22,r11,r12,r22) \
{r11 = (x11) - (y11);\
 r12 = (x12) - (y12);\
 r22 = (x22) - (y22);}

#define M2X2_SUBTRACT_SYM(x11,x12,x22,y11,y12,y22,r11,r12,r22) \
{r11 = (x11) - (y11);\
 r12 = (x12) - (y12);\
 r22 = (x22) - (y22);}

#define M2X2XV2(x11,x12,x21,x22,v1,v2,r1,r2) \
{r1 = ((x11) * (v1)) + ((x12) * (v2));\
 r2 = ((x21) * (v1)) + ((x22) * (v2));}

#define M2X2_SUB_OP_NORM(x11,x12,x21,x22,v1,v2,u1,u2,c,r11,r12,r21,r22)\
{r11 = (x11) - (((v1) * (u1)) / (c));\
 r12 = (x12) - (((v1) * (u2)) / (c));\
 r21 = (x21) - (((v2) * (u1)) / (c));\
 r22 = (x22) - (((v2) * (u2)) / (c));}

#define M2X2_SUB_OP_NORM_SYM(x11,x12,x22,v1,v2,c,r11,r12,r22)\
{r11 = (x11) - (((v1) * (v1)) / (c));\
 r12 = (x12) - (((v1) * (v2)) / (c));\
 r22 = (x22) - (((v2) * (v2)) / (c));}

#define V2DOT(v1,v2,u1,u2) (((v1) * (u1)) + ((v2) * (u2)))

#define PI         3.1415926535897932384626433832795
#define PI_180     0.0174532925199432957692369076849
#define PI_OVER_2  (PI/2.0)
#define PI2        (PI*2.0)

#define PAIRS_CONVERGE 0
#define STEEPEST_DESCENT 1
#define INITIALIZE 0
#define UPDATE 1

#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))

/* BOOLEAN */

#define BOOLEAN unsigned char
#define	TRUE	1
#define	FALSE	0

/*** IMAGE DATATYPES ***/

#define IMAGE unsigned char
#define RLI_TYPE short

#define NIL       -1
#define UNDEFINED  0

/***** RETURN STATUS CODES *************************************************/

#define EXCEEDED_FRAG_LIMIT	1
#define LINE_LIST_OVERFLOW	2
#define FET_OVERFLOW		3

#ifdef SEQUENT
#define RED    0
#define GREEN  1
#define BLUE   2
#define YELLOW 3
#else
#define RED    1
#define GREEN  2
#define BLUE   3
#define YELLOW 4
#endif

typedef struct pair {
  int i;
  int j;
  float distance;
  float length;
  } PAIR;

typedef struct tableau {
  REAL e_match;
  REAL e_fit;
  REAL e_om;
  REAL e_s;
  REAL e_match_best;
  int undefinedp;
  int tests;
  int moves;
  ABCUVK cor;
  STRANS trans;
  int n;          /* Number of data lines */
  PR *pairs;
  int m;          /* Number of model lines */
  ML *curr_mls;
  ML *init_mls;
  REAL mls_cum_len;
  int mlprs_n;
  PAIR *mlpairs;
  } TABLEAU;

/*
  MODEL_INFO
    Used to record information about the models.
*/

typedef struct {
  REAL x_range;
  REAL y_range;
  REAL theta_range;
  int number_model_lines;
  int current_model_line;
  int number_data_lines;
  int current_trial;
  REAL best_e_match;
  LINE3D *lines3d;
  ML *lines2d;
  TABLEAU *tab;
  TABLEAU *best_tableau;
  } MODEL_INFO;
